# Generated by Django 4.2.7 on 2024-02-05 04:39

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("database", "0029_userrequests"),
    ]

    operations = [
        migrations.AddField(
            model_name="conversation",
            name="slug",
            field=models.CharField(blank=True, default=None, max_length=200, null=True),
        ),
        migrations.AddField(
            model_name="conversation",
            name="title",
            field=models.CharField(blank=True, default=None, max_length=200, null=True),
        ),
    ]
